@@ -87,8 +87,6 @@ extern "C" {
 using namespace std;
 
 static void set_details_callback (void (*func) (gpointer), gpointer data);
-static void set_operation_label (const char *label);
-static void set_operation_callback (void (*func) (gpointer), gpointer data);
 
 static void get_package_infos_in_background (GList *packages);
 
@@ -168,7 +166,18 @@ static GList *install_sections = NULL;
 static GList *upgradeable_packages = NULL;
 static GList *installed_packages = NULL;
 static GList *search_result_packages = NULL;
-static gboolean package_list_ready = false;
+
+
+enum package_list_state {
+  pkg_list_unknown,
+  pkg_list_retrieving,
+  pkg_list_ready,
+};
+
+static package_list_state pkg_list_state = pkg_list_unknown;
+
+#define package_list_ready (pkg_list_state == pkg_list_ready)
+
 
 static int cur_section_rank;
 static char *cur_section_name;
@@ -182,39 +191,29 @@ set_current_view (view *v)
 
   if (v->id == MAIN_VIEW)
     {
-      enable_update_all (false);
       enable_refresh (false);
-      enable_search (false);
-      enable_sort (false);
+      prevent_updating ();
     }
   else if (v->id == UNINSTALL_APPLICATIONS_VIEW)
     {
-      enable_update_all (false);
       enable_refresh (false);
-      enable_search (true);
-      enable_sort (true);
+      allow_updating ();
     }
   else if (v->id == INSTALL_SECTION_VIEW
            || v->id == SEARCH_RESULTS_VIEW)
     {
-      enable_update_all (false);
       enable_refresh (true);
-      enable_search (true);
-      enable_sort (true);
+      allow_updating ();
     }
   else if (v->id == INSTALL_APPLICATIONS_VIEW)
     {
-      enable_update_all (false);
       enable_refresh (true);
-      enable_search (true);
-      enable_sort (install_sections && !install_sections->next);
+      allow_updating ();
     }
   else if (v->id == UPGRADE_APPLICATIONS_VIEW)
     {
-      enable_update_all (package_list_ready && upgradeable_packages);
       enable_refresh (true);
-      enable_search (true);
-      enable_sort (true);
+      allow_updating ();
     }
 
   cur_view_struct = v;
@@ -237,11 +236,13 @@ show_view (view *v)
     }
 
   set_details_callback (NULL, NULL);
-  set_operation_label (NULL);
-  set_operation_callback (NULL, NULL);
 
   allow_updating ();
 
+  /* Reset global path when changing among views */
+  if (v != cur_view_struct)
+    reset_global_target_path ();
+
   v->cur_view = v->maker (v);
   v->dirty = false;
 
@@ -254,7 +255,6 @@ show_view (view *v)
 static void
 show_upgrade_applications_view_and_refresh_callback (GtkWidget *btn, gpointer data)
 {
-  package_list_ready = false;
   show_check_for_updates_view ();
 
   if (is_idle ())
@@ -355,8 +355,6 @@ make_main_view (view *v)
 
   get_package_infos_in_background (NULL);
 
-  enable_search (false);
-
   prevent_updating ();
 
   return view;
@@ -477,6 +475,7 @@ section_info::section_info ()
   ref_count = 1;
   rank = 1;
   name = NULL;
+  untranslated_name = NULL;
   packages = NULL;
 }
 
@@ -564,6 +563,10 @@ nicify_section_name (const char *name)
   if (*name == '\0')
     return "-";
 
+  // "Other" is not an good section for ALL rank
+  if (!g_strcmp0 (name, "other"))
+    return NULL;
+
   // try the translation from the official text domain.
   char buf[200];
   snprintf (buf, 200, "ai_category_%s", name);
@@ -608,8 +611,13 @@ static section_info *
 create_section_info (GList **list_ptr,
 		     int rank, const char *name)
 {
+  const char *untranslated_name = NULL;
+
   if (name)
-    name = nicify_section_name (canonicalize_section_name (name));
+    {
+      untranslated_name = canonicalize_section_name (name);
+      name = nicify_section_name (name);
+    }
 
   if (!name)
     {
@@ -617,20 +625,28 @@ create_section_info (GList **list_ptr,
 	 "normal" section.  Move it to the "other" rank in that case.
       */
       if (rank == SECTION_RANK_NORMAL)
-	rank = SECTION_RANK_OTHER;
+        {
+          if (untranslated_name && !strcmp (untranslated_name, "hidden"))
+            rank = SECTION_RANK_HIDDEN;
+          else
+            rank = SECTION_RANK_OTHER;
+        }
 
       if (rank == SECTION_RANK_ALL)
 	name = _("ai_category_all");
+      else if (rank == SECTION_RANK_HIDDEN)
+        name = "hidden";
       else if (rank == SECTION_RANK_OTHER)
 	name = _("ai_category_other");
     }
 
   section_info *si = find_section_info (list_ptr, rank, name);
-  
+
   if (si == NULL)
     {
       si = new section_info;
       si->rank = rank;
+      si->untranslated_name = untranslated_name;
       si->name = name;
       si->packages = NULL;
       if (list_ptr)
@@ -955,7 +971,7 @@ get_package_list_reply (int cmd, apt_proto_decoder *dec, void *data)
 	all_si->unref ();
     }
 
-  package_list_ready = true;
+  pkg_list_state = pkg_list_ready;
 
   /* Refresh view after sorting only if not in the main view */
   sort_all_packages (cur_view_struct != &main_view);
@@ -994,7 +1010,7 @@ get_package_list_with_cont (void (*cont) (void *data), void *data)
   /* Mark package list as not ready and cancel the package info
      getting in the background before freeing the list
   */
-  package_list_ready = false;
+  pkg_list_state = pkg_list_retrieving;
   get_package_infos_in_background (NULL);
   free_all_packages ();
 
@@ -1407,6 +1423,40 @@ install_package_flow (package_info *pi)
     }
 }
 
+static GList *
+update_all_get_upgradeable_packages (void)
+{
+  GList *packages = NULL;
+  GList *item;
+
+  /* Look for the OS package */
+  for (item = upgradeable_packages; item; item = g_list_next (item))
+    {
+      package_info *pi = (package_info *)item->data;
+
+      if (pi->flags & pkgflag_system_update)
+        {
+          /* Ensure only the OS package is included */
+          if (packages != NULL)
+            {
+              g_list_free (packages);
+              packages = NULL;
+            }
+          packages = g_list_append (packages, pi);
+
+          break;
+        }
+      else
+        {
+          /* Just append it if not an OS update */
+          packages = g_list_prepend (packages, pi);
+        }
+    }
+
+  /* Return reversed list */
+  return g_list_reverse (packages);
+}
+
 static void
 update_all_packages_flow_end (int n_successful, void *data)
 {
@@ -1420,7 +1470,7 @@ update_all_packages_flow ()
 {
   if (start_interaction_flow ())
     {
-      GList *packages_list = g_list_copy (upgradeable_packages);
+      GList *packages_list = update_all_get_upgradeable_packages ();
       install_packages (packages_list,
 			INSTALL_TYPE_UPGRADE_ALL_PACKAGES,
 			false,  NULL, NULL,
@@ -1428,12 +1478,6 @@ update_all_packages_flow ()
     }
 }
 
-static void
-install_operation_callback (gpointer data)
-{
-  install_package_flow ((package_info *)data);
-}
-
 static void
 ignore_package_info (package_info *pi, void *data, bool changed)
 {
@@ -1445,14 +1489,12 @@ available_package_selected (package_info *pi)
   if (pi)
     {
       set_details_callback (available_package_details, pi);
-      set_operation_callback (install_operation_callback, pi);
       pi->ref ();
       get_package_info (pi, true, ignore_package_info, NULL);
     }
   else
     {
       set_details_callback (NULL, NULL);
-      set_operation_callback (NULL, NULL);
     }
 }
 
@@ -1487,27 +1529,13 @@ uninstall_package_flow (package_info *pi)
     }
 }
 
-static void
-uninstall_operation_callback (gpointer data)
-{
-  uninstall_package_flow ((package_info *)data);
-}
-
 void
 installed_package_selected (package_info *pi)
 {
   if (pi)
-    {
-      set_details_callback (installed_package_details, pi);
-      set_operation_callback (uninstall_operation_callback, pi);
-      set_operation_label (_("ai_me_package_uninstall"));
-    }
+    set_details_callback (installed_package_details, pi);
   else
-    {
-      set_details_callback (NULL, NULL);
-      set_operation_callback (NULL, NULL);
-      set_operation_label (_("ai_tb_uninstall"));
-    }
+    set_details_callback (NULL, NULL);
 }
 
 static void
@@ -1524,28 +1552,20 @@ make_install_section_view (view *v)
 {
   GtkWidget *view;
 
-  set_operation_label (_("ai_tb_install"));
-
   section_info *si = find_section_info (&install_sections,
 					cur_section_rank, cur_section_name);
 
-  view =
-    make_global_package_list (si? si->packages : NULL,
-			      false,
-			      (package_list_ready
-			       ? _("ai_li_no_applications_available")
-			       : NULL),
-			      _("ai_me_cs_install"),
-			      available_package_selected, 
-			      available_package_activated);
-
+  view = make_install_apps_package_list (v->window,
+                                         si? si->packages : NULL,
+                                         package_list_ready,
+                                         available_package_selected,
+                                         available_package_activated);
   if (package_list_ready)
-    gtk_widget_show_all (view);
+    gtk_widget_show (view);
 
   if (si)
     get_package_infos_in_background (si->packages);
 
-  enable_search (true);
   enable_refresh (true);
 
   return view;
@@ -1684,20 +1704,18 @@ make_install_applications_view (view *v)
 
   check_catalogues ();
 
-  set_operation_label (_("ai_tb_install"));
-
   if (install_sections && install_sections->next == NULL)
     {
       section_info *si = (section_info *)install_sections->data;
       view =
-	make_global_package_list (si->packages,
-				  false,
-				  (package_list_ready
-				   ? _("ai_li_no_applications_available")
-				   : NULL),
-				  _("ai_me_cs_install"),
-				  available_package_selected,
-				  available_package_activated);
+        make_install_apps_package_list (v->window,
+                                        ((si->rank == SECTION_RANK_HIDDEN)
+                                         ? NULL
+                                         : si->packages),
+                                        package_list_ready,
+                                        available_package_selected,
+                                        available_package_activated);
+
       get_package_infos_in_background (si->packages);
     }
   else
@@ -1706,7 +1724,7 @@ make_install_applications_view (view *v)
     }
 
   if (package_list_ready)
-    gtk_widget_show_all (view);
+    gtk_widget_show (view);
 
   maybe_refresh_package_cache_without_user ();
 
@@ -1732,6 +1750,13 @@ show_check_for_updates_view ()
   show_view (&upgrade_applications_view);
 }
 
+void
+show_install_applications_view ()
+{
+  show_view (&install_applications_view);
+  set_current_view (&install_applications_view);
+}
+
 static void
 update_seen_updates_file (void)
 {
@@ -1763,20 +1788,14 @@ make_upgrade_applications_view (view *v)
 
   check_catalogues ();
 
-  set_operation_label (_("ai_tb_update"));
-
-  view =
-    make_global_package_list (upgradeable_packages,
-			      false,
-			      (package_list_ready
-			       ? _("ai_li_no_updates_available")
-			       : NULL),
-			      _("ai_me_cs_update"),
-			      available_package_selected,
-			      available_package_activated);
-
+  view = make_upgrade_apps_package_list (v->window,
+                                         upgradeable_packages,
+                                         package_list_ready,
+                                         package_list_ready && upgradeable_packages,
+                                         available_package_selected,
+                                         available_package_activated);
   if (package_list_ready)
-    gtk_widget_show_all (view);
+    gtk_widget_show (view);
 
   get_package_infos_in_background (upgradeable_packages);
 
@@ -1784,7 +1803,6 @@ make_upgrade_applications_view (view *v)
       && hildon_window_get_is_topmost (HILDON_WINDOW (get_main_window ())))
     {
       update_seen_updates_file ();
-      enable_update_all (true);
     }
 
   // maybe_refresh_package_cache_without_user ();
@@ -1796,21 +1814,14 @@ GtkWidget *
 make_uninstall_applications_view (view *v)
 {
   GtkWidget *view;
-
-  set_operation_label (_("ai_tb_uninstall"));
-
-  view = make_global_package_list (installed_packages,
-				   true,
-				   (package_list_ready
-				    ? _("ai_li_no_installed_applications")
-				    : NULL),
-				   _("ai_me_cs_uninstall"),
-				   installed_package_selected,
-				   installed_package_activated);
+  view = make_uninstall_apps_package_list (v->window,
+                                           installed_packages,
+                                           package_list_ready,
+                                           installed_package_selected,
+                                           installed_package_activated);
   if (package_list_ready)
-    gtk_widget_show_all (view);
+    gtk_widget_show (view);
 
-  enable_search (true);
   enable_refresh (false);
 
   return view;
@@ -1824,46 +1835,60 @@ make_search_results_view (view *v)
   if (v->parent == &install_applications_view
       || v->parent == &upgrade_applications_view)
     {
-      set_operation_label (v->parent == &install_applications_view
-			   ? _("ai_tb_install")
-			   : _("ai_tb_update"));
+      if (v->parent == &install_applications_view)
+        {
+          view = make_install_apps_package_list (v->window,
+                                                 search_result_packages,
+                                                 FALSE,
+                                                 available_package_selected,
+                                                 available_package_activated);
+        }
+      else
+        {
+          view = make_upgrade_apps_package_list (v->window,
+                                                 search_result_packages,
+                                                 FALSE,
+                                                 package_list_ready && search_result_packages,
+                                                 available_package_selected,
+                                                 available_package_activated);
+        }
 
-      view = make_global_package_list (search_result_packages,
-				       false,
-				       NULL,
-				       (v->parent == &install_applications_view
-					? _("ai_me_cs_install")
-					: _("ai_me_cs_update")),
-				       available_package_selected,
-				       available_package_activated);
       get_package_infos_in_background (search_result_packages);
     }
   else
     {
-      set_operation_label (_("ai_tb_uninstall"));
-
-      view = make_global_package_list (search_result_packages,
-				       true,
-				       NULL,
-				       _("ai_me_cs_uninstall"),
-				       installed_package_selected,
-				       installed_package_activated);
+      view = make_uninstall_apps_package_list (v->window,
+                                               search_result_packages,
+                                               FALSE,
+                                               installed_package_selected,
+                                               installed_package_activated);
     }
-  gtk_widget_show_all (view);
+  gtk_widget_show (view);
 
-  enable_search (true);
   enable_refresh (true);
 
   return view;
 }
 
+/*
+ * Return true if all the search_words match with the package name
+ */
+static bool
+match_pattern (const char *pkg_name, char **search_words)
+{
+  int i;
+
+  for (i = 0; search_words[i] != NULL; i++)
+    if (!strcasestr (pkg_name, search_words[i]))
+      return false;
+  return true;
+}
+
 static void
 search_package_list (GList **result,
                      GList *packages, const char *pattern, bool installed)
 {
   gchar **words;
-  gboolean found;
-  int i;
 
   if (!(words = g_strsplit (pattern, " ", 0)))
     return; /* pattern is empty */
@@ -1871,18 +1896,12 @@ search_package_list (GList **result,
   while (packages)
     {
       package_info *pi = (package_info *)packages->data;
-      found = FALSE;
 
-      for (i = 0; words[i] != NULL; i++)
-        if (strcasestr (pi->get_display_name (installed), words[i]))
-          found = TRUE;
-        else
-          {
-            found = FALSE;
-            break;
-          }
-
-      if (found)
+      /* Insert only packages that match with search pattern and also
+       * either has an installed version (are installed) or are not hidden
+       */
+      if (match_pattern (pi->get_display_name (installed), words)
+          && (pi->installed_version || !package_is_hidden (pi)))
         {
           pi->ref ();
           *result = g_list_append (*result, pi);
@@ -1890,6 +1909,7 @@ search_package_list (GList **result,
 
       packages = packages->next;
     }
+
   g_strfreev (words);
 }
 
@@ -1972,7 +1992,15 @@ search_packages_reply (int cmd, apt_proto_decoder *dec, void *data)
 	  if (install_sections)
 	    {
 	      section_info *si = (section_info *)install_sections->data;
-	      find_in_package_list (&result, si->packages, name);
+
+              // Not packages in section HIDDEN
+              // But if the section is ALL then the package shouldn't be hidden
+              if (si->rank == SECTION_RANK_HIDDEN
+                  || (si->rank == SECTION_RANK_ALL
+                      && (!info->installed_version && package_is_hidden (info))))
+                ;
+              else
+                find_in_package_list (&result, si->packages, name);
 	    }
 	}
       else if (parent == &upgrade_applications_view)
@@ -2055,7 +2083,10 @@ search_packages (const char *pattern, bool in_descriptions)
 	  if (install_sections)
 	    {
 	      section_info *si = (section_info *)install_sections->data;
-	      search_package_list (&result, si->packages, pattern, false);
+
+              // Just for precaution: avoid the HIDDEN rank
+              if (si->rank != SECTION_RANK_HIDDEN)
+                search_package_list (&result, si->packages, pattern, false);
 	    }
 	}
       else if (parent == &upgrade_applications_view)
@@ -2252,34 +2283,6 @@ set_details_callback (void (*func) (gpointer), gpointer data)
   details_func = func;
 }
 
-static void (*operation_func) (gpointer);
-static gpointer operation_data;
-static const char *operation_label;
-
-void
-do_current_operation ()
-{
-  if (operation_func)
-    operation_func (operation_data);
-}
-
-static void
-set_operation_label (const char *label)
-{
-  if (label == NULL)
-    label = _("ai_tb_install");
-
-  operation_label = label;
-}
-
-static void
-set_operation_callback (void (*func) (gpointer), gpointer data)
-{
-  operation_data = data;
-  operation_func = func;
-  set_operation_label (operation_label);
-}
-
 /* Reinstalling the packages form the recently restored backup.
  */
 
@@ -2387,6 +2390,9 @@ rp_end (int n_successful, void *data)
     xexp_free (backup);
 
   end_interaction_flow ();
+
+  /* Make sure packages list is initialized after this */
+  maybe_init_packages_list ();
 }
 
 /* INSTALL_FROM_FILE_FLOW
@@ -2436,6 +2442,9 @@ static void
 iff_end (bool success, void *unused)
 {
   end_interaction_flow ();
+
+  /* Make sure packages list is initialized after this */
+  maybe_init_packages_list ();
 }
 
 static void
@@ -2451,9 +2460,37 @@ view_set_dirty (view *v)
     v->dirty = true;
 }
 
+static void
+is_topmost_cb (GtkWidget *widget, GParamSpec *arg, gpointer data)
+{
+  g_return_if_fail(widget != NULL && HILDON_IS_WINDOW(widget));
+
+  HildonWindow *window = HILDON_WINDOW (widget);
+
+  set_current_view ((view *) data);
+  g_debug ("the top view is %d", cur_view_struct->id);
+
+  /* Update the seen-updates file if the window is top most again and
+     the "Check for updates" view is currently selected */
+  if (package_list_ready && hildon_window_get_is_topmost (window) &&
+      (cur_view_struct == &upgrade_applications_view))
+    {
+      update_seen_updates_file ();
+    }
+
+  if (cur_view_struct->dirty)
+    show_view (cur_view_struct);
+}
+
 static void
 reset_view (view *v)
 {
+  /* Disconnect window's signal handlers using the user_data parameter */
+  g_signal_handlers_disconnect_by_func (G_OBJECT (v->window),
+                                        (gpointer) is_topmost_cb,
+                                        v);
+
+  /* Set NULL values */
   v->window = NULL;
   v->cur_view = NULL;
   //  cur_view_struct = v->parent;
@@ -2495,36 +2532,6 @@ main_window_realized (GtkWidget* widget, gpointer unused)
 	      "hildon-application-manager");
 }
 
-static void
-cancel_download (void *unused)
-{
-  cancel_apt_worker ();
-}
-
-static void
-apt_status_callback (int cmd, apt_proto_decoder *dec, void *unused)
-{
-  if (dec == NULL)
-    return;
-
-  int op = dec->decode_int ();
-  int already = dec->decode_int ();
-  int total = dec->decode_int ();
-
-  if (total > 0)
-    {
-      if (op == op_downloading)
-	{
-	  set_entertainment_download_fun (op, already, total);
-	  set_entertainment_cancel (cancel_download, NULL);
-	}
-      else
-	{
-	  set_entertainment_fun (NULL, op, already, total);
-	}
-    }
-}
-
 GtkWindow *
 get_main_window ()
 {
@@ -2533,28 +2540,6 @@ get_main_window ()
   return main_window;
 }
 
-static void
-is_topmost_cb (GtkWidget *widget, GParamSpec *arg, gpointer data)
-{
-  g_return_if_fail(widget != NULL && HILDON_IS_WINDOW(widget));
-
-  HildonWindow *window = HILDON_WINDOW (widget);
-
-  set_current_view ((view *) data);
-  g_debug ("the top view is %d", cur_view_struct->id);
-
-  /* Update the seen-updates file if the window is top most again and
-     the "Check for updates" view is currently selected */
-  if (package_list_ready && hildon_window_get_is_topmost (window) &&
-      (cur_view_struct == &upgrade_applications_view))
-    {
-      update_seen_updates_file ();
-    }
-
-  if (cur_view_struct->dirty)
-    show_view (cur_view_struct);
-}
-
 static gboolean
 key_event (GtkWidget *widget,
 	   GdkEventKey *event,
@@ -2694,6 +2679,15 @@ with_initialized_packages (void (*cont) (void *data), void *data)
     }
 }
 
+void maybe_init_packages_list (void)
+{
+  if (initial_packages_available || (pkg_list_state != pkg_list_unknown))
+    return;
+
+  get_package_list_with_cont (notice_initial_packages_available, NULL);
+  save_backup_data ();
+}
+
 osso_context_t *
 get_osso_context ()
 {
@@ -2703,7 +2697,6 @@ get_osso_context ()
 int
 main (int argc, char **argv)
 {
-  const char *apt_worker_prog = "/usr/libexec/apt-worker";
   bool show = true;
 
   if (argc > 1 && !strcmp (argv[1], "--no-show"))
@@ -2714,7 +2707,7 @@ main (int argc, char **argv)
     }
   if (argc > 1)
     {
-      apt_worker_prog = argv[1];
+      set_apt_worker_cmd (argv[1]);
       argc--;
       argv++;
     }
@@ -2763,16 +2756,11 @@ main (int argc, char **argv)
 
   create_menu ();
 
-  if (!start_apt_worker (apt_worker_prog))
-    what_the_fock_p ();
-
-  apt_worker_set_status_callback (apt_status_callback, NULL);
-
-  get_package_list_with_cont (notice_initial_packages_available, NULL);
-  save_backup_data ();
-
   if (show)
-    present_main_window ();
+    {
+      maybe_init_packages_list ();
+      present_main_window ();
+    }
 
   gtk_main ();
 }
